<?php
/**
 * @file
 *
 * Class for handling background processes.
 */

/**
 * BackgroundProcess class.
 */
class BackgroundProcess {
  public $handle;
  public $connection = NULL;
  public $service_host;
  public $service_group;

  /**
   * Constructor.
   *
   * @param type $handle
   *   Handle to use. Optional; leave out for auto-handle.
   */
  function __construct($handle = NULL) {
    $this->handle = $handle;
    $this->service_group = variable_get('background_process_default_service_group', NULL);
  }

  /**
   * Destructor. Clean up.
   */
  function __destruct() {
    $this->cleanUp();
  }

  /**
   * Clean up. Remove process from DB if applicable.
   */
  function cleanUp() {
    if ($this->connection === FALSE) {
      background_process_remove_process($this->handle);
    }
    unset($this->connection);
  }

  /**
   * Start background process
   *
   * Calls the service handler through http passing function arguments as serialized data
   * Be aware that the callback will run in a new request
   *
   * @global string $base_url
   *   Base URL for this Drupal request
   *
   * @param $callback
   *   Function to call.
   * @param $args
   *   Array containg arguments to pass on to the callback.
   * @return mixed
   *   TRUE on success, NULL on failure, FALSE on handle locked.
   */
  function start($callback, $args) {
    // Generate token
    $token = background_process_generate_handle($callback);

    if (!isset($this->handle)) {
      $this->handle = $token;
    }
    $handle = $this->handle;

    module_invoke_all('background_process_pre_execute', $handle, $callback, $args, $token);

    // Preliminary select to avoid unnecessary write-attempt
    if (background_process_get_process($handle)) {
      watchdog('bg_process', 'Will not attempt to set handle %handle for callback %callback', array('%handle' => $handle, '%callback' => $callback), WATCHDOG_ERROR);
      return FALSE;
    }

    // "Lock" handle
    if (!background_process_set_process($handle, $callback, NULL, $args, $token)) {
      // If this happens, we might have an md5 clash
      watchdog('bg_process', 'Could not set handle %handle for callback %callback', array('%handle' => $handle, '%callback' => $callback), WATCHDOG_ERROR);
      return FALSE;
    }
    // Initialize progress stats
    progress_remove_progress($handle);

    $this->connection = FALSE;

    // Find service group if a service host is not explicitly specified.
    if (!$this->service_host) {
      if (!$this->service_group) {
        $this->service_group = variable_get('background_process_default_service_group', NULL);
      }
      if ($this->service_group) {
        $service_groups = variable_get('background_process_service_groups', array());
        if (isset($service_groups[$this->service_group])) {
          $service_group = $service_groups[$this->service_group];
          $this->service_host = module_invoke($service_group['module'], 'service_group', 'service_host', $service_group);
        }
      }
    }

    // Find service host to use
    if (!$this->service_host) {
      $this->service_host = variable_get('background_process_default_service_host', NULL);
    }
    if ($this->service_host) {
      $service_hosts = variable_get('background_process_service_hosts', array());
      if (isset($service_hosts[$this->service_host])) {
        $service_host = $service_hosts[$this->service_host];
      }
    }

    // Default fallback host
    if (!$service_host) {
      if (!$service_host) {
        global $base_url;
        $service_host = array(
          'base_url' => $base_url,
        );
      }
    }

    // Generate service URL
    $path = '/background-process/start/' . urlencode($handle) . '/' . urlencode($token);
    $url = parse_url($service_host['base_url']);
    $port = isset($url['port']) ? $url['port'] : 80;
    $host = isset($url['host']) ? $url['host'] : NULL;
    $http_host = isset($service_host['http_host']) ? $service_host['http_host'] : $host . (isset($url['port']) ? ':' . $url['port'] : '');

    // What? No Host?!?
    if (empty($host)) {
      $this->cleanUp();
      watchdog('bg_process', 'No host for %handle %callback', array('%handle' => $handle, '%callback' => $callback), WATCHDOG_ERROR);
      // Throw exception here instead?
      return NULL;
    }

    // Setup variable to pass to background function call
    $content  = "";
    $length = strlen($content);
    $type = "application/x-www-form-urlencoded";

    // Use current session, etc., by resending request-headers
    $headers = _background_process_request_headers();
    $headers = _background_process_filter_headers($headers);
    $headers = $headers ? implode("\r\n", $headers) . "\r\n" : '';

    db_query("UPDATE {background_process} SET `start` = %f, service_host = '%s' WHERE handle = '%s'", microtime(TRUE), $this->service_host, $handle);

    // Call the service through HTTP POST
    watchdog('bg_process', "Invoke:  $handle ($host:$port$path) ($http_host)", array(), WATCHDOG_DEBUG);
    $fp = fsockopen($host, $port, $errno, $errstr, variable_get('background_process_connection_timeout', BACKGROUND_PROCESS_CONNECTION_TIMEOUT));
    if ($fp) {

      $out = "GET {$path} HTTP/1.1\r\n";
      $out .= "Host: {$http_host}\r\n";
      $out .= "Connection: Close\r\n";
      $out .= "Content-Type: $type\r\n";
      $out .= "Content-Length: $length\r\n";
      $out .= $headers;
      $out .= "\r\n";
      $out .= $content;
      stream_set_write_buffer($fp, 0);
      $written = fwrite($fp, $out);
      watchdog('bg_process', 'Written: %written', array('%written' => $written), WATCHDOG_DEBUG);

      stream_set_blocking($fp, FALSE);
      stream_set_timeout($fp, variable_get('background_process_stream_timeout', BACKGROUND_PROCESS_STREAM_TIMEOUT));
      $info = stream_get_meta_data($fp);
      $this->connection = $fp;
      return TRUE;
    }
    else {
      $this->cleanUp();
      watchdog('bg_process', 'Could not call service %handle for callback %callback', array('%handle' => $handle, '%callback' => $callback), WATCHDOG_ERROR);
      // Throw exception here instead?
      return NULL;
    }
  }
}
